/*

=head1 Tcl method invocation functions
=head2 Description
The functions in this file are responsible for actually executing
the messages sent to objects. They find out which method to call
for a given object/message and call this method with the appropriate
attributes. There are also routines that handle error reporting and
debugging.

	Copyright (c) 1995 by softWorks.  All Rights Reserved.
	File:   method.c
	Author: Richard Schwaninger
	Date:   30/10/1995
	System: ithron.softworks.co.at
	RCS:    $Id: c_method.c,v 1.1 1996/08/30 08:48:44 /SW/app Exp /SW/app $

*/


#define _method_c

#include "oo.h"


PRIVATE int OO_EvalMethod(Tcl_Interp *,char *, char *,
                         char *, int,char **);

extern int OO_Eval(Tcl_Interp *,char *,char *,int, char **);
extern int OO_EvalArgv(Tcl_Interp *,...);


PRIVATE int
OO_FindMethod (
   Tcl_Interp *interp,
   char *Name,
      char *This,
      char *Msg,
      CallPtr call
   )
{
   char MetaClass[50]="Meta-";
   char *FindFunc;
   char *kind;
   int ret;
   int ac;
   char **av;

   /* printf("FindMethod for %s:%s, %s\n",Name,This,Msg);*/
   /* fflush(stdout); */

   kind=GET_GLOB(Name,"Kind");
   if(kind==NULL) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,"Class '",Name,"' is not known (",
                       This," ",Msg,").",NULL);
      return(TCL_ERROR);
   }
      
   strcat(MetaClass,kind);
   FindFunc=GET_GLOB(MetaClass,"Find");
   if(FindFunc==NULL) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,"Class '",Name,"' for object '",
                       This,"' has no 'Find' function.",NULL);
      return(TCL_ERROR);
   }
   ret=OO_EvalArgv(interp,FindFunc,Name,This,Msg,NULL);
   if(ret!=TCL_OK) {
      return(TCL_ERROR);
   }
   
   /* this returns {} when nothing was found */
   
   if(interp->result[0]=='\0') {
      return(TCL_OK);
   }
   
   if(Tcl_SplitList(interp,interp->result,&ac,&av)!=TCL_OK) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,"retvalue of FindFunc for class '",
                       Name,"' (oject '",This,
                       "') was not splittable.",
                       NULL);
      return(TCL_ERROR);
   }
   SET_CONTEXT(call,av[0],av[1],av[2]);
   free(av);
   return(TCL_OK);
}



PRIVATE void OO_Debug(
   Tcl_Interp *interp,
      int what,
      int Status,
      char *Class,
      char *This,
      char *Msg,
      int argc,
      char **argv
   )
{
   static int level=0;
   int flag;
   ooPtr odata=Tcl_GetAssocData(interp,"OO",NULL);

   /*
          odata->OO_DebugFlag,what);
*/
   if(what==0) {
      ++level;
      ++odata->OO_DebugTop;
   } else {
      if(what==1) --level;
      --odata->OO_DebugTop;
   }
   
   if(odata->OO_DebugFlag<0) {
      Tcl_VarEval(interp,odata->OO_DebugFunc," ",Msg," ",This,NULL);
      return;
   }
   
   if(odata->OO_DebugFlag>0) {
      flag=odata->OO_DebugFlag;
      odata->OO_DebugFlag=0;
      if((flag==1)&&(what==0)) {
         int i;
         char buf[500];
         Tcl_DString ds;
         Tcl_DStringInit(&ds);
         sprintf(buf,"M-%d: %s %s",level,This,Msg);
         Tcl_DStringAppend(&ds,buf,-1);
         for(i=0;i<argc;i++) {
            Tcl_DStringAppendElement(&ds,argv[i]);
         }
         Tcl_DStringAppend(&ds,"\n",-1);
         Tcl_Write(odata->TraceChannel,Tcl_DStringValue(&ds),-1);
         Tcl_DStringFree(&ds);
      }
      odata->OO_DebugFlag=flag;
   }
}



PRIVATE int 
OO_Error (
   Tcl_Interp *interp,
      char *Class,
      char *This,
      int argc,
      char **argv
    ) 
{
   int result;
   char *err=strdup(interp->result);

   /* printf("Error: Class=%s ('%s') [%d]\n",Class,interp->result,argc); */

   result=OO_EvalMethod(interp,Class,This,"Error",1,&err);
   free(err);
   return(result);
}


PRIVATE int 
OO_Unknown (
   Tcl_Interp *interp,
      char *Class,
      int argc,
      char **argv
   )
{
   printf("unknown not yet done\n");
   return(TCL_BREAK);
}



static
int OO_HandleResult(
   Tcl_Interp *interp,
      int retval,                  /* old call return value */
      int unknown,                /* to call Unkown=1 */
      char *Class,
      int argc,
      char **argv
   )
{
   Tcl_DString einfo;
   Tcl_DString result;
   int ret;
   
   Tcl_DStringInit(&result);
   Tcl_DStringInit(&einfo);

   Tcl_DStringGetResult(interp,&result);   
   Tcl_DStringAppend(&einfo,Tcl_GetVar2(interp,"errorInfo",
                                        NULL,TCL_GLOBAL_ONLY),-1);


   if((retval==TCL_BREAK)&&unknown) {
      /* method does not exist */
      ret=OO_EvalMethod(interp,Class,argv[0],"Unknown",
                        argc-1,argv+1);
      if(ret==TCL_OK) {
         Tcl_DStringFree(&result);
         Tcl_DStringFree(&einfo);
         return(TCL_OK);
      }

      Tcl_ResetResult(interp);
      Tcl_AddErrorInfo(interp,Tcl_DStringValue(&einfo));
      Tcl_SetResult(interp,Tcl_DStringValue(&result),TCL_VOLATILE);
      /* fall thru */
   }
   /* an error occured */

   ret=OO_Error(interp,Class,argv[0],argc,argv);
   if(ret==TCL_BREAK) {
      Tcl_ResetResult(interp);
      Tcl_AddErrorInfo(interp,Tcl_DStringValue(&einfo));
      Tcl_SetResult(interp,Tcl_DStringValue(&result),TCL_VOLATILE);
      ret=TCL_ERROR;
   }

   Tcl_DStringFree(&result);
   Tcl_DStringFree(&einfo);
   return ret;
}


/*
=head1 OO_ThatCmd - call some method in parent
=head2 Description
This function is responsible for calling a message in the parent
context. It notes the current context of the object, finds the
method corresponding to the given message in one of the parents
and executes that message with the given arguments.

If more than one parent has a method for the given message then 
they are all executed (in an arbitrary order). If a C<TCL_ERROR> or
a C<TCL_BREAK> code is returned from a call then this chaining is
aborted and in the first case an error is returned.

This function implements the tcl command C<OO_That> and in turn the
call sequence C<$That msg ...>

Here is an example:

	Method bla { a1 a2 } {
		$That bla $a1
		puts "$a2"
	}

=head2 Arguments
This command has the standard tcl call syntax.
=head2 Return Value
Returns a standard tcl result code and sets C<interp-E<gt>result>.

*/

TCL_COMMAND(OO_That Msg args,OO_ThatCmd) 
{
   int i,ac;
   char **av;
   char *parents;
   int ret=TCL_OK;
   
   ooPtr odata=Tcl_GetAssocData(interp,"OO",NULL);
   /* printf("THAT: %s (%d)\n",argv[1],argc);*/

   if(argc<2) {
      Tcl_SetResult(interp,"wrong # of args to OO_That",TCL_STATIC);
      return(TCL_ERROR);
   }

   if((parents=GET_GLOB(odata->CallTop->Context,"Parents"))==NULL) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,"Object '",odata->CallTop->Id,
                       "' has no parents for 'That'.",NULL);
      return(TCL_OK);
   }

   if(Tcl_SplitList(interp,parents,&ac,&av)!=TCL_OK) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,"Split error for parents of '",
                       odata->CallTop->Id,"' for 'That'.",NULL);
      return(TCL_OK);
   }
   for(i=0;i<ac;i++) {
      ret=OO_EvalMethod(interp,av[i],odata->CallTop->Id,argv[1],argc-2,argv+2);
      if(ret==TCL_ERROR) {
         free(av);
         return(OO_HandleResult(interp,TCL_BREAK,1
                                ,odata->CallTop->Context,argc,argv));
      }
      if(ret==TCL_BREAK) {
         continue;
      }
      if(ret==TCL_OK) {
         /*
*/
         break;
      }
   }
   free(av);
   return(TCL_OK);
}


TCL_COMMAND(OO_Super Parent Msg args,OO_SuperCmd) 
{
   int ret;
   
   ooPtr odata=Tcl_GetAssocData(interp,"OO",NULL);
   /* printf("Super: %s %s (%d)\n",argv[1],argv[2],argc);*/

   if(argc<3) {
      Tcl_SetResult(interp,"wrong # of args to OO_Super",TCL_STATIC);
      return(TCL_ERROR);
   }
   
   ret=OO_EvalMethod(interp,argv[1],odata->CallTop->Id,argv[2],argc-3,argv+3);
   if(ret==TCL_OK) {
      return(ret);
   }

   return(OO_HandleResult(interp,ret,1,odata->CallTop->Context,argc,argv));
}


TCL_COMMAND(OO_Self Msg args,OO_SelfCmd) 
{
   int ret;
   
   ooPtr odata=Tcl_GetAssocData(interp,"OO",NULL);
   if(argc<2) {
      Tcl_SetResult(interp,"wrong # of args to OO_Self",TCL_STATIC);
      return(TCL_ERROR);
   }
   /* printf("Self: %s (%d)\n",argv[1],argc);*/

   
   ret=OO_EvalMethod(interp,odata->CallTop->Context,odata->CallTop->Id,
                     argv[1],argc-2,argv+2);
   if(ret==TCL_OK) {
      return(ret);
   }

   return(OO_HandleResult(interp,ret,1,odata->CallTop->Context,argc,argv));
}


/*
=head1 OO_EvalMethod - evaluate method command
=head2 Description
This function takes the given arguments, find's a method to call for the
give message, builds up the corresponding tcl string and then evaluates
this string. If an error happens then the corresponding error message
is tried and called, otherwise the error is returned. 

If the method cannot be found anywhere in the inheritance chain then
the C<Unknown> message is sent to the current object (if that does not
exist too then an error happens).

Just before and just after the method invocation the object debugger 
entry is called (if enabled) so we can check and modify arguments.
=head2 Arguments.

=over 3
=item interp
Pointer to current interpreter
=item Class
Name of class for the current object
=item This
Object name for the current object
=item Msg
The message signature that was sent to the current object
=item argc
number of arguments for this message
=item argv
list of actual arguments for this message
=back
=head2 Return Value
This function returns a I<Tcl_Result>-code, either obtained by calling
C<OO_Eval()> or from the method lookup. It returns C<TCL_BREAK> to 
signify something special, but I don't know what.
*/

PRIVATE int
OO_EvalMethod (
   Tcl_Interp *interp,
      char *Class,
      char *This,
      char *Msg,
      int argc,
      char **argv
)
{
   CallPtr call;
   int ret;
   ooPtr odata=Tcl_GetAssocData(interp,"OO",NULL);
   
   /* printf("EvalMethod %s %s %s\n",Class,This,Msg); */
   PUSH_CONTEXT(call);

   if(OO_FindMethod(interp,Class,This,Msg,call)!=TCL_OK) {
      POP_CONTEXT(call);
      return(TCL_ERROR);
   }
   if(call->Call==NULL) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,"No method '",
                       Msg,"' for '",This,"' ('",Class,"').",
                       NULL);
      POP_CONTEXT(call);
      return(TCL_BREAK);
   }

   /* printf("before eval of %s %s (%d)\n",call->Call,call->Id,argc); */
   OO_Debug(interp,0,-1,Class,This,Msg,argc,argv);
   ret=OO_Eval(interp,call->Call,call->Id,argc,argv);
   OO_Debug(interp,1,ret,Class,This,Msg,argc,argv);   
   
   if(ret==TCL_ERROR) {
      Tcl_DString ds;
      int i;

      Tcl_DStringInit(&ds);
      Tcl_DStringAppend(&ds,"\nMethod",-1);
      Tcl_DStringAppendElement(&ds,This);
      Tcl_DStringAppendElement(&ds,Msg);
      for(i=0;i<argc;i++) {
         Tcl_DStringAppendElement(&ds,argv[i]);
      }
      Tcl_AddErrorInfo(interp,Tcl_DStringValue(&ds));
      Tcl_DStringFree(&ds);
   }
   POP_CONTEXT(call);
   return(ret);
}

   

/*
=head1 OO_EvalMethodCmd - method evaluator
=head2 Description
This routine is the procedure visible in the tcl interpreter. It does
some simple housekeeping and allows for debugging before calling the
main evaluator routine C<OO_EvalMethod>.

=head2 Arguments
This command has the standard tcl call syntax

=head2 Return Value
The function returns either C<TCL_OK> or C<TCL_ERROR> to indicate
success or failure of the call.

*/

TCL_COMMAND(OO_EvalMethodCmd Class Obj Msg args,OO_EvalMethodCmd) 
{
   char *Class=(char *)Data;
   int ret;
   ooPtr odata=Tcl_GetAssocData(interp,"OO",NULL);
   
   /* printf("Method %s %s %s\n",Class,argv[0],argv[1]); */
   if(argc<2) {
      Tcl_SetResult(interp,"wrong # of args to OO_EvalMethodCmd",
                    TCL_STATIC);
      return(TCL_ERROR);
   }
   
   odata->OO_ErrorFlag=0;
   odata->OO_DebugTop=0;
   ret=OO_EvalMethod(interp,Class,argv[0],argv[1],
                     argc-2,argv+2);
   
   if(ret==TCL_OK) {
      return(TCL_OK);
   }

   ret=OO_HandleResult(interp,ret,1,Class,argc,argv);
   return(ret);
}

   


TCL_COMMAND(OO_MaybeEvalMethod Class Obj Msg args,OO_MaybeEvalMethodCmd) 
{
   int ret;
   ooPtr odata=Tcl_GetAssocData(interp,"OO",NULL);
   
   if(argc<4) {
      Tcl_SetResult(interp,"wrong # of args to OO_MaybeEvalMethodCmd",
                    TCL_STATIC);
      return(TCL_ERROR);
   }
   

   /* printf("maybe: '%s' '%s' '%s' (%d)\n",argv[1],argv[2],argv[3],argc); */
   odata->OO_ErrorFlag=0;
   odata->OO_DebugTop=0;
   ret=OO_EvalMethod(interp,argv[1],argv[2],argv[3],
                     argc-4,argv+4);
   /* printf("retval is %d,%s\n",ret,interp->result); */
   if((ret==TCL_OK)||(ret==TCL_BREAK)) {
      return(TCL_OK);
   }
   return(OO_HandleResult(interp,ret,0,argv[1],argc,argv));
   
}

   



void OO_MethodDelCmd(
   ClientData clientData
)
{
   free(clientData);
}

   


TCL_COMMAND(OO_Method name ,OO_DoMethodCmd) 
{

   /* argv[1]=Class, argv[2]=Id, */
   char *classname;
   if((argc!=3)||(argv[1][0]=='\0')||(argv[2][0]=='\0')) {
      Tcl_SetResult(interp,"wrong # of args to OO_Method",TCL_STATIC);
      return(TCL_ERROR);
   }
   
   classname=strdup(argv[1]);
   Tcl_CreateCommand(interp,argv[2],OO_EvalMethodCmd,
                     classname,OO_MethodDelCmd);
   Tcl_SetResult(interp,argv[2],TCL_VOLATILE);
   return(TCL_OK);
}

   



TCL_COMMAND(bla,OO_MethodTraceCmd) 
{
   int mode;
   ooPtr odata=Tcl_GetAssocData(interp,"OO",NULL);

   if(odata->TraceChannel==NULL) {
      odata->TraceChannel=Tcl_GetChannel(interp,"stdout",&mode);
      if((odata->TraceChannel==NULL)&&((mode&TCL_WRITABLE)==0)) {
         Tcl_SetResult(interp,"stdout not open for write",TCL_STATIC);
         return(TCL_ERROR);
      }
   }
      
   if(strcmp(argv[1],"on")==0) {
      if(argc==3) {
         odata->OO_DebugFlag=-1;
         odata->OO_DebugFunc=strdup(argv[2]);
         return(TCL_OK);
      }
      odata->OO_DebugFlag=1;
   } else {
      if(odata->OO_DebugFlag==-1) {
         free(odata->OO_DebugFunc);
         odata->OO_DebugFunc=NULL;
      }
      
      odata->OO_DebugFlag=0;

   }
   return(TCL_OK);
}


#undef _method_c */

/* EOF */
















































































































